from cloud_governance.common.logger.init_logger import logger
from cloud_governance.aws.tag_non_cluster.remove_non_cluster_tags import RemoveNonClusterTags
from cloud_governance.aws.tag_non_cluster.tag_non_cluster_resources import TagNonClusterResources
from cloud_governance.aws.tag_non_cluster.update_na_tag_resources import UpdateNATags


def tag_non_cluster_resource(mandatory_tags: dict, region: str, tag_operation: str = 'read'):
    if tag_operation == 'update':
        action = 'Tag'
        dry_run = 'no'
    else:
        action = 'Scan'
        dry_run = 'yes'

    tag_non_cluster_resources = TagNonClusterResources(input_tags=mandatory_tags, region=region, dry_run=dry_run)

    func_resource_list = [tag_non_cluster_resources.non_cluster_update_ec2,
                          tag_non_cluster_resources.update_volumes,
                          tag_non_cluster_resources.update_ami,
                          tag_non_cluster_resources.update_snapshots,
                          ]
    logger.info(f"{action} {len(func_resource_list)} non-cluster resources  in region {region}:")
    for func in func_resource_list:
        func()


def remove_tag_non_cluster_resource(mandatory_tags: dict, region: str, dry_run: str = 'yes'):
    remove_tag_non_cluster_resources = RemoveNonClusterTags(input_tags=mandatory_tags, region=region, dry_run=dry_run)

    func_resource_list = [remove_tag_non_cluster_resources.non_cluster_update_ec2,
                          remove_tag_non_cluster_resources.update_volumes,
                          remove_tag_non_cluster_resources.update_ami,
                          remove_tag_non_cluster_resources.update_snapshots,
                          ]
    if dry_run == "no":
        action = 'Tag'
    else:
        action = 'Scan'
    logger.info(f"{action} {len(func_resource_list)} non-cluster resources  in region {region}:")
    for func in func_resource_list:
        func()


def tag_na_resources(file_name: str, file_path: str = '/tmp/', tag_operation: str = 'read', region: str = 'us-east-2'):
    """
    This method generate the NA tag file and update tags of each region
    @param file_path:
    @param tag_operation:
    @param region:
    @param file_name:
    @return:
    """
    file_name = f'{file_path}{file_name}'
    update_na_tags = UpdateNATags(region=region, file_name=file_name)
    if tag_operation == 'read':
        logger.info('Generating the CSV file')
        update_na_tags.create_csv()
    elif tag_operation == 'update':
        update_na_tags.update_tags()





from cloud_governance.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class ZombieNatGateways(NonClusterZombiePolicy):

    def __init__(self):
        super().__init__()

    def __check_nat_gateway_in_routes(self, nat_gateway_id: str):
        route_tables = self._ec2_client.describe_route_tables()['RouteTables']
        nat_gateway_found = False
        for route_table in route_tables:
            for route in route_table.get('Routes'):
                if route.get('NatGatewayId') == nat_gateway_id:
                    nat_gateway_found = True
        return nat_gateway_found

    def run(self):
        """
        This method return zombie NatGateways, delete if dry_run no
        @return:
        """
        nat_gateways = self._ec2_operations.get_nat_gateways()
        zombie_nat_gateways_data = []
        for nat_gateway in nat_gateways:
            nat_gateway_id = nat_gateway.get('NatGatewayId')
            tags = nat_gateway.get('Tags')
            gateway_unused = False
            if not self._check_live_cluster_tag(tags=tags):
                if nat_gateway.get('State') == 'available':
                    if not self.__check_nat_gateway_in_routes(nat_gateway_id=nat_gateway_id):
                        gateway_unused = True
                        unused_days = self._get_resource_last_used_days(tags=tags)
                        zombie_nat_gateway = self._check_resource_and_delete(resource_name='NatGateway',
                                                                             resource_id='NatGatewayId',
                                                                             resource_type='CreateNatGateway',
                                                                             resource=nat_gateway,
                                                                             empty_days=unused_days,
                                                                             days_to_delete_resource=self.DAYS_TO_DELETE_RESOURCE, tags=tags)
                        if zombie_nat_gateway:
                            zombie_nat_gateways_data.append([nat_gateway_id, self._get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                             zombie_nat_gateway.get('VpcId'), self._get_policy_value(tags=tags), unused_days])
                    else:
                        unused_days = 0
                    self._update_resource_tags(resource_id=nat_gateway_id, tags=tags, left_out_days=unused_days, resource_left_out=gateway_unused)
        return zombie_nat_gateways_data




from cloud_governance.common.clouds.ibm.account.ibm_account import IBMAccount


class ClassicOperations:
    """"
    This class is for IBM classic operations - BareMetal, Virtual Machines
    """

    def __init__(self):
        self._sl_client = IBMAccount().get_sl_client()

    def __filter_tag_names(self, tag_references: list):
        """
        This method filter tag names from the tagReferences
        @param tags:
        @return:
        """
        tags = []
        if tag_references:
            for tag in tag_references:
                tags.append(tag.get('tag')['name'].strip())
        return tags

    def get_hardware_ids(self):
        """
        this method list all hardwares ( bare-metal machines) in ibm classic infrastructure devices
        @return:
        """
        hardware_mask = "mask[id, hostname]"
        hardware_ids = self._sl_client.call('Account', 'getHardware', mask=hardware_mask, iter=True)
        return hardware_ids

    def get_hardware_data(self, hardware_id: str):
        """
        This method return hardware_data from the hardware-id
        hardware --> bare-metal machine
        @param hardware_id:
        @return:
        """
        mask = 'mask[billingItem[orderItem[order[userRecord[username]]]]]'
        hardware_data = self._sl_client.call('SoftLayer_Hardware_Server', 'getObject', id=hardware_id, mask=mask)
        return hardware_data

    def get_hardware_tags(self, hardware_id: str):
        """
        This method returns tags of hardware ( bare-metal machines )
        @param hardware_id:
        @return:
        """
        tags = self._sl_client.call('SoftLayer_Hardware_Server', 'getTagReferences', id=hardware_id)
        return self.__filter_tag_names(tag_references=tags)

    def get_virtual_machine_ids(self):
        """
        this method list all hardwares ( bare-metal machines) in ibm classic infrastructure devices
        @return:
        """
        vm_mask = "mask[id, hostname]"
        vm_ids = self._sl_client.call('Account', 'getVirtualGuests', mask=vm_mask, iter=True)
        return vm_ids

    def get_virtual_machine_data(self, vm_id: str):
        """
        This method return virtual machine from the hardware-id
        @param vm_id:
        @return:
        """
        mask = 'mask[billingItem[orderItem[order[userRecord[username]]]]]'
        vm_data = self._sl_client.call('SoftLayer_Virtual_Guest', 'getObject', id=vm_id, mask=mask)
        return vm_data

    def get_virtual_machine_tags(self, vm_id: str):
        """
        This method returns tags of hardware ( bare-metal machines )
        @param vm_id:
        @return:
        """
        tags = self._sl_client.call('SoftLayer_Virtual_Guest', 'getTagReferences', id=vm_id)
        return self.__filter_tag_names(tag_references=tags)




from cloud_governance.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class ZombieElasticIps(NonClusterZombiePolicy):

    def __init__(self):
        super().__init__()

    def run(self):
        """
        This method return zombie elastic_ip's and delete if dry_run no
        @return:
        """
        addresses = self._ec2_operations.get_elastic_ips()
        zombie_addresses = []
        for address in addresses:
            ip_no_used = False
            tags = address.get('Tags')
            if not self._check_live_cluster_tag(tags=tags):
                if not address.get('NetworkInterfaceId'):
                    ip_no_used = True
                    unused_days = self._get_resource_last_used_days(tags=tags)
                    zombie_eip = self._check_resource_and_delete(resource_name='ElasticIp',
                                                                 resource_id='AllocationId',
                                                                 resource_type='AllocateAddress',
                                                                 resource=address,
                                                                 empty_days=unused_days,
                                                                 days_to_delete_resource=self.DAYS_TO_DELETE_RESOURCE, tags=tags)
                    if zombie_eip:
                        zombie_addresses.append([address.get('AllocationId'), self._get_tag_name_from_tags(tags=tags),
                                                 self._get_tag_name_from_tags(tags=tags, tag_name='User'), address.get('PublicIp'),
                                                 self._get_policy_value(tags=tags), unused_days])
                else:
                    unused_days = 0
                self._update_resource_tags(resource_id=address.get('AllocationId'), tags=tags, left_out_days=unused_days, resource_left_out=ip_no_used)
        return zombie_addresses




from cloud_governance.common.logger.init_logger import logger
from cloud_governance.ibm.tagging.tagging_operations import TaggingOperations


class TagBareMetal(TaggingOperations):
    """
    This method tag IBM bare-metal machines
    """
    def __init__(self):
        super().__init__()

    def get_hardware_username(self, hardware_id: str):
        """
        This method returns the hardware_username form the order details
        @param hardware_id:
        @return:
        """
        hardware_data = self._classic_operations.get_hardware_data(hardware_id=hardware_id)
        return hardware_data.get('billingItem').get('orderItem').get('order').get('userRecord').get('username'), f'{hardware_data.get("hostname")}.{hardware_data.get("domain")}'

    def tag_remove_hardware(self, user_tags: list, hardware_tags: list, hardware_id: str, hardware_name: str):
        """
        This method removes the hardware tags
        @param user_tags:
        @param hardware_tags:
        @param hardware_id:
        @param hardware_name:
        @return:
        """
        remove_hardware_tags = self._filter_remove_tags(user_tags=user_tags, resource_tags=hardware_tags)
        if self._tag_remove_name:
            remove_hardware_tags = [self._tag_remove_name]
        if remove_hardware_tags:
            try:
                response = self.softlayer_operation(softlayer_name='SoftLayer_Hardware_Server', softlayer_method='removeTags', resource_id=hardware_id, tags=','.join(remove_hardware_tags))
                if response:
                    logger.info(f'Tags are Removed to the hardware: {hardware_id} - {hardware_name} : count: {len(remove_hardware_tags)} : {remove_hardware_tags}')
                else:
                    logger.info(f'Tags are not Removed to the Hardware: {hardware_id} - {hardware_name}, something might fail')
            except Exception as err:
                logger.info(f'{err}')
        return remove_hardware_tags

    def tag_update_hardware(self, user_tags: list, hardware_tags: list, hardware_id: str, hardware_name: str):
        """
        This method sets the hardware tags
        @param user_tags:
        @param hardware_tags:
        @param hardware_id:
        @param hardware_name:
        @return:
        """
        add_hardware_tags = []
        if self._tag_custom:
            add_hardware_tags.extend(self._tag_custom)
        add_hardware_tags.extend(self._filter_common_tags(user_tags=user_tags, resource_tags=hardware_tags))
        if add_hardware_tags:
            if self._tag_operation == 'update':
                add_hardware_tags.extend(hardware_tags)
                add_hardware_tags = list(set(add_hardware_tags))
                try:
                    response = self.softlayer_operation(softlayer_name='SoftLayer_Hardware_Server', softlayer_method='setTags', resource_id=hardware_id, tags=','.join(add_hardware_tags))
                    if response:
                        logger.info(f'Tags are added to the hardware: {hardware_id} - {hardware_name} : count: {len(add_hardware_tags)} : {add_hardware_tags}')
                    else:
                        logger.info(f'Tags are not added to the hardware: {hardware_id} - {hardware_name}, something might fail')
                except Exception as err:
                    logger.info(f'{err}')
        return add_hardware_tags

    def tag_hardware(self, hardware_id: str):
        """
        This method perform tag operations - update, remove read
        @param hardware_id:
        @return:
        """
        username, hardware_name = self.get_hardware_username(hardware_id=hardware_id)
        hardware_tags = self._classic_operations.get_hardware_tags(hardware_id=hardware_id)
        user_tags = self._ibm_client.get_user_tags_from_gsheet(username=username)
        if self._tag_operation == 'remove':
            tags = self.tag_remove_hardware(user_tags, hardware_tags, hardware_id, hardware_name)
        else:
            tags = self.tag_update_hardware(user_tags, hardware_tags, hardware_id, hardware_name)
        return tags

    def run(self, hardware_id: str = ''):
        """
        This method tag hardware ( bare-metals ) from the user tags from the gsheet
        @return:
        """
        response = []
        if hardware_id:
            response = self.tag_hardware(hardware_id=hardware_id)
        else:
            hardware_ids = self._classic_operations.get_hardware_ids()
            for hardware_id in hardware_ids:
                response.append({hardware_id.get('hostname'): self.tag_hardware(hardware_id=hardware_id.get('id'))})
        return response



import os

import SoftLayer
import pandas as pd


from cloud_governance.common.google_drive.google_drive_operations import GoogleDriveOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch


class IBMAccount:
    """
    This class contains IBM softlayer client and methods
    """

    def __init__(self):
        self.__API_USERNAME = os.environ.get('IBM_API_USERNAME', '')
        self.__API_KEY = os.environ.get('IBM_API_KEY', '')
        self.__sl_client = SoftLayer.Client(username=self.__API_USERNAME, api_key=self.__API_KEY)
        self.__account = os.environ.get('account', '')
        self.__gsheet_id = os.environ.get('SPREADSHEET_ID', '')
        self.__gsheet_client = GoogleDriveOperations()
        self.__ldap_host_name = os.environ.get('LDAP_HOST_NAME', '')
        self.__ldap = LdapSearch(ldap_host_name=self.__ldap_host_name)

    def get_sl_client(self):
        """
        This method return the softlayer client
        @return:
        """
        return self.__sl_client

    def __organise_user_tags(self, tags: dict):
        """
        This method organise the tags from the gsheet
        @param tags:
        @return:
        """
        user_tags = []
        for tag, value in tags.items():
            if value.strip():
                value = value.replace('/', '-')
                user_tags.append(f'{tag.strip().lower()}:{value.strip().lower()}')
        return user_tags

    def get_user_tags_from_gsheet(self, username: str, file_path: str = '/tmp/'):
        """
        This method return the user tags from the gsheet
        @param username:
        @param file_path:
        @return:
        """
        file_name = os.path.join(file_path, f'{self.__account}.csv')
        if not os.path.exists(file_name):
            self.__gsheet_client.download_spreadsheet(spreadsheet_id=self.__gsheet_id, sheet_name=self.__account, file_path=file_path)
        df = pd.read_csv(file_name)
        df.fillna('', inplace=True)
        df.set_index('User', inplace=True)
        tags = dict(df.loc[username])
        tags['user'] = username.split('_')[1].split('@')[0]
        for key in list(tags.keys()):
            if key.startswith('_'):
                tags.pop(key)
        ldap_data = self.__ldap.get_user_details(user_name=tags['user'])
        if ldap_data:
            tags['Owner'] = ldap_data['FullName']
            tags['Manager'] = ldap_data['managerName']
        return self.__organise_user_tags(tags)



import operator
from operator import ge

import boto3
from botocore.exceptions import ClientError

from cloud_governance.common.logger.init_logger import logger
from cloud_governance.aws.zombie_non_cluster.run_zombie_non_cluster_policies import NonClusterZombiePolicy


class EmptyBuckets(NonClusterZombiePolicy):

    def __init__(self):
        super().__init__()

    def run(self):
        """
        This method return all Empty buckets and delete if dry_run no
        @return:
        """
        return self.__delete_empty_bucket()

    def __delete_empty_bucket(self):
        """
        This method delete the empty bucket more than 7 days
        @return:
        """
        empty_buckets = []
        buckets = self._s3_client.list_buckets()['Buckets']
        for bucket in buckets:
            bucket_empty = False
            empty_days = 0
            bucket_name = bucket.get('Name')
            try:
                try:
                    bucket_tags = self._s3_client.get_bucket_tagging(Bucket=bucket_name)
                    tags = bucket_tags.get('TagSet')
                except ClientError:
                    tags = []
                bucket_data = self._s3_client.list_objects_v2(Bucket=bucket_name)
                if not bucket_data.get('Contents'):
                    if not self._check_live_cluster_tag(tags=tags):
                        if not self._get_tag_name_from_tags(tags=tags, tag_name='Name'):
                            tags.append({'Key': 'Name', 'Value': bucket_name})
                        empty_days = self._get_resource_last_used_days(tags=tags)
                        bucket_empty = True
                        if not self._get_tag_name_from_tags(tags=tags, tag_name='User'):
                            region = self._s3_client.get_bucket_location(Bucket=bucket_name)['LocationConstraint']
                            self._cloudtrail.set_cloudtrail(region_name=region)
                        empty_bucket = self._check_resource_and_delete(resource_name='S3 Bucket', resource_id='Name', resource_type='CreateBucket', resource=bucket, empty_days=empty_days, days_to_delete_resource=self.DAYS_TO_DELETE_RESOURCE, tags=tags)
                        if empty_bucket:
                            empty_buckets.append([bucket.get('Name'), self._get_tag_name_from_tags(tags=tags, tag_name='User'), str(bucket.get('CreationDate')), str(empty_days), self._get_policy_value(tags=tags)])
                else:
                    empty_days = 0
                self._update_resource_tags(resource_id=bucket_name, tags=tags, left_out_days=empty_days, resource_left_out=bucket_empty)
            except Exception as err:
                logger.info(f'{err}, {bucket.get("Name")}')
        return empty_buckets



import json
import os


class MailMessage:
    RESTRICTION = 'Don\'t replay this email, in case of issue update us in slack channel #perf-dept-public-clouds.'

    def __init__(self):
        self.account = os.environ.get('account', '')
        self.policy = os.environ.get('policy', '')
        self.region = os.environ.get('AWS_DEFAULT_REGION', '')

    def ec2_stop(self, name: str, days: int, image_id: str, delete_instance_days: int, instance_name: str,
                 resource_id: str, stopped_time: str, ec2_type: str):
        subject = f'cloud-governance alert: ec2-stop'
        content = 'If you want that cloud-governance will not stop it add Policy=Not_Delete tag to your instance.'
        message = f'Cloud-governance will stop it in the {delete_instance_days-days} days.'
        if image_id != '':
            content = f'You can find a image of the deleted image under AMI: {image_id}'
            message = f'This instance will be deleted due to it was stopped more than {delete_instance_days} days.'
        body = f"""
Greetings {name},

Instance: {instance_name}: {resource_id}( InstanceType:{ec2_type} ) in {self.region} on AWS account: {self.account} was stopped on {stopped_time}, it stopped more than {days} days.  
{message}
{content}

{self.RESTRICTION}

Best regards,
Cloud-governance Team""".strip()
        return subject, body

    def ec2_idle(self, name: str, days: int, notification_days: int, stop_days: int, instance_name: str,
                 resource_id: str, ec2_type: str):
        subject = f'cloud-governance alert:  ec2-idle'
        if days == notification_days:
            cause = f'This instance will be stopped in {stop_days-days} days'
            content = 'If you want that cloud-governance will not stop it add Policy=Not_Delete tag to your instance.'
        else:
            cause = f'This instance will be stopped.'
            content = 'In future cloud-governance will not stop it add Policy=Not_Delete tag to your instance'
        body = f"""
Greetings {name},

Instance: {instance_name}: {resource_id} ( InstanceType:{ec2_type} ) in {self.region} on AWS account: {self.account} is idle more than {days} days.
{cause}
{content}
If you already added the Policy=Not_Delete tag ignore this mail.

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body

    def iam_user_add_tags(self, name: str, user: str, spreadsheet_id: str):
        subject = f'cloud-governance alert: Missing tags in AWS IAM User'
        body = f"""
Hi {name},

{os.environ.get('account')} IAM User: {user} has missing tags 
Please add the tags in the spreadsheet: https://docs.google.com/spreadsheets/d/{spreadsheet_id}.
If you already filled the tags, please ignore the mail.

{self.RESTRICTION}

Best Regards
Cloud-governance Team""".strip()
        return subject, body

    def aws_user_over_usage_cost(self, user: str, usage_cost: int, name: str, user_usage: int):
        """
        This method send subject, body to over usage cost
        @return:
        """
        subject = f'cloud-governance alert: last week cost over usage {usage_cost}$ per user'
        body = f"""
Hi {name},
Your AWS user:{user} in account:{os.environ.get('account')} cost was {user_usage}$ in the last week. 
Please verify that you are using all the resources.
Please verify that you are using all the resources in attached {user}_resource.json file.

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body

    def resource_message(self, name: str, days: int, notification_days: int, delete_days: int, resource_name: str, resource_id: str, resource_type: str, resources: list = []):
        """
        This method prepare mail message based on resource_type and return subject, body
        @param resources:
        @param name:
        @param days:
        @param notification_days:
        @param delete_days:
        @param resource_name:
        @param resource_id:
        @param resource_type:
        @return:
        """
        resource_type = resource_type.capitalize()
        reason = self.policy.split('_')[-1]
        if 'empty' in self.policy:
            reason = 'empty'
        if 'zombie' in self.policy:
            reason = 'Unused'
        subject = f'cloud-governance alert: {self.policy}'
        if days == notification_days:
            cause = f'This {resource_type} will be deleted in  {delete_days-days} days'
            content = f'If you want that cloud-governance will not delete it add Policy=Not_Delete/skip tag to your {resource_type}s.'
        else:
            cause = f'This {resource_type} will be deleted due to it was {reason} more than {delete_days} days.'
            content = f'In future cloud-governance will not delete your {resource_type} add Policy=Not_Delete/skip tag to your {resource_type}s'
        extra_data = ''
        if resources:
            extra_data = f'Cluster Undeleted Resources: {sorted(resources)}'
        body = f"""
Greetings {name},

{resource_type}: {resource_name}: {resource_id} in {self.region} on AWS account: {self.account} is {reason} more than {days} days.
{cause}
{content}
If you already added the Policy=Not_Delete/skip tag ignore this mail.

{extra_data}

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body

    def zombie_cluster_mail_message(self, name: str, days: int, notification_days: int, resource_name: str, delete_days: int):
        """
        This method prepare mail message based on resource_type and return subject, body
        @param delete_days:
        @param name:
        @param days:
        @param notification_days:
        @param resource_name:
        @return:
        """
        subject = f'Cloud-governance policy alert: {self.policy}'
        if days == notification_days:
            cause = f'Cloud-governance will delete those {self.policy}s in the  {delete_days-days} days.'
        else:
            cause = f'Those {self.policy}s will be deleted.'
        body = f"""
Hi {name},

There are zombie resources that related to your terminated cluster: {resource_name}, region: {self.region}, AWS account: {self.account}
You can find your {self.policy}s in the attached file: {resource_name.replace('/', '-')}.json
{cause}

{self.RESTRICTION}

Best Regards,
Cloud-governance Team""".strip()
        return subject, body



import datetime
import os
import smtplib
from email.message import EmailMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from cloud_governance.common.clouds.aws.s3.s3_operations import S3Operations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.logger.init_logger import logger

# https://github.com/redhat-performance/quads/blob/master/quads/tools/postman.py


class Postfix:
    """
    This class send mail using the postfix configuration
    install postfix and configure on machine
    $ dnf install postfix -y
    $ vi /etc/postfix/main.cf
    relayhost = [smtp.corp.redhat.com]
    inet_interfaces = all
    inet_protocols = all
    relay_domains = rdu2.scalelab.redhat.com, redhat.com
    $ systemctl restart postfix
    """

    def __init__(self):
        self.reply_to = os.environ.get('REPLY_TO', 'dev-null@redhat.com')
        self.__es_host = os.environ.get('es_host', '')
        self.__policy = os.environ.get('policy', '')
        self.__es_port = os.environ.get('es_port', '')
        self.__account = os.environ.get('account', '')
        self.__policy_output = os.environ.get('policy_output', '')
        self.bucket_name, self.key = self.get_bucketname()
        self.__es_index = 'cloud-governance-mail-messages'
        if self.__es_host:
            self.__es_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port)
        if self.__policy_output:
            self.__s3_operations = S3Operations(region_name='us-east-1')

    def get_bucketname(self):
        bucket_name = ''
        key = 'logs'
        if 's3' in self.__policy_output.lower():
            targets = self.__policy_output.split('/')
            bucket_name = targets[2]
            key = targets[3]
        else:
            bucket_name = self.__policy_output
        return bucket_name, key

    def send_email_postfix(self, subject: str, to: str, cc: list, content: str, **kwargs):
        msg = MIMEMultipart('alternative')
        msg["Subject"] = subject
        msg["From"] = "%s <%s>" % (
            'cloud-governance',
            "@".join(["noreply-cloud-governance", 'redhat.com']),
        )
        msg["To"] = "@".join([to, 'redhat.com'])
        msg["Cc"] = ",".join(cc)
        # msg.add_header("Reply-To", self.reply_to)
        # msg.add_header("User-Agent", self.reply_to)
        if kwargs.get('filename'):
            attachment = MIMEText(open(kwargs['filename']).read())
            attachment.add_header('Content-Disposition', 'attachment',
                                  filename=kwargs['filename'].split('/')[-1])
            msg.attach(attachment)
        msg.attach(MIMEText(content))
        email_string = msg.as_string()
        email_host = 'localhost'
        try:
            with smtplib.SMTP(email_host) as s:
                try:
                    logger.debug(email_string)
                    s.send_message(msg)
                    logger.info(f'Mail sent successfully to {to}@redhat.com')
                    if kwargs.get('filename'):
                        file_name = kwargs['filename'].split('/')[-1]
                        date_key = datetime.datetime.now().strftime("%Y%m%d%H")
                        self.__s3_operations.upload_file(file_name_path=kwargs['filename'],
                                                         bucket=self.bucket_name, key=f'{self.key}/{self.__policy}/{date_key}',
                                                         upload_file=file_name)
                        s3_path = f'{self.__policy_output}/logs/{self.__policy}/{date_key}/{file_name}'
                        content += f'\n\nresource_file_path: s3://{s3_path}\n\n'
                    data = {'Policy': self.__policy, 'To': to, 'Cc': cc, 'Message': content, 'Account': self.__account.upper()}
                    if kwargs.get('resource_id'):
                        data['resource_id'] = kwargs['resource_id']
                    if self.__es_host:
                        self.__es_operations.upload_to_elasticsearch(data=data, index=self.__es_index)
                except smtplib.SMTPException as ex:
                    logger.info(f'Error while sending mail, {ex}')
                    return False
            return True
        except Exception as err:
            logger.info(f'Some error occurred, {err}')
            return False



from datetime import datetime, timedelta

import boto3


class CloudWatchOperations:
    """
    This class perform the cloudwatch operations
    methods
    1. get_metric_data
    """

    def __init__(self, region: str = 'us-east-2'):
        self._region = region
        self.cloudwatch_client = boto3.client('cloudwatch', region_name=self._region)

    def _create_metric_lists(self, resource_id: str, resource_type: str, namespace: str, metric_names: dict, statistic: str):
        """
        This method create the metrics from metric resources
        @param resource_id:
        @param resource_type:
        @param namespace:
        @param metric_names:
        @param statistic:
        @return:
        """
        metric_lists = []
        for index, (metric_name, unit) in enumerate(metric_names.items()):
            metric_lists.append(
                {
                    'Id': f'metric{index}',
                    'MetricStat': {
                        'Metric': {
                            'Namespace': namespace,
                            'MetricName': metric_name,
                            'Dimensions': [{'Name': resource_type, 'Value': resource_id}]
                        },
                        'Period': 86400,
                        'Stat': statistic,
                        'Unit': unit
                    },
                }
            )
        return metric_lists

    def get_metric_data(self, start_time: datetime, end_time: datetime, resource_id: str, resource_type: str,
                        namespace: str, metric_names: dict, statistic: str):
        """
        This method returns metrics of the specified resource and metrics
        @param start_time:
        @param end_time:
        @param resource_id:
        @param resource_type:
        @param namespace:
        @param metric_names:
        @param statistic:
        @return:
        """
        metric_lists = self._create_metric_lists(resource_id=resource_id, resource_type=resource_type,
                                                 namespace=namespace, metric_names=metric_names, statistic=statistic)
        return self.cloudwatch_client.get_metric_data(StartTime=start_time, EndTime=end_time,
                                                      MetricDataQueries=metric_lists)




import os
import typeguard
from ast import literal_eval  # str to dict
import boto3  # regions

from cloud_governance.aws.cost_expenditure.cost_report_policies import CostReportPolicies
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp, logger
from cloud_governance.aws.tag_cluster.run_tag_cluster_resouces import tag_cluster_resource, remove_cluster_resources_tags
from cloud_governance.aws.tag_non_cluster.run_tag_non_cluster_resources import tag_non_cluster_resource, remove_tag_non_cluster_resource, tag_na_resources
from cloud_governance.aws.tag_user.run_tag_iam_user import tag_iam_user, run_validate_iam_user_tags
from cloud_governance.aws.zombie_cluster.run_zombie_cluster_resources import zombie_cluster_resource
from cloud_governance.gitleaks.gitleaks import GitLeaks
from cloud_governance.ibm.ibm_operations.tag_resources import TagResources
from cloud_governance.main.es_uploader import ESUploader
from cloud_governance.common.clouds.aws.s3.s3_operations import S3Operations
from cloud_governance.aws.zombie_cluster.validate_zombies import ValidateZombies
from cloud_governance.aws.zombie_non_cluster.zombie_non_cluster_polices import ZombieNonClusterPolicies

# env tests
# os.environ['AWS_DEFAULT_REGION'] = 'us-east-2'
# os.environ['AWS_DEFAULT_REGION'] = 'all'
# os.environ['policy'] = 'zombie_cluster_resource'
# os.environ['validate_type'] = 'tags'
# os.environ['user_tags'] = "['Budget', 'User', 'Owner', 'Manager', 'Environment', 'Project']"
# os.environ['cost_metric'] = ''
# os.environ['start_date'] = ''
# os.environ['end_date'] = ''
# os.environ['granularity'] = ''
# os.environ['policy'] = 'ec2_untag'
# os.environ['policy'] = 'zombie_cluster_resource'
# os.environ['dry_run'] = 'yes'
# os.environ['tag_operation'] = 'read'
# os.environ['service_type'] = 'ec2_zombie_resource_service'
# os.environ['service_type'] = 'iam_zombie_resource_service'
# os.environ['service_type'] = 's3_zombie_resource_service'
# os.environ['resource'] = 'zombie_cluster_elastic_ip'
# os.environ['resource'] = 'zombie_cluster_nat_gateway'
# os.environ['cluster_tag'] = ''
# os.environ['cluster_tag'] = ''
# os.environ['policy_output'] = 's3://bucket_name/logs'
# os.environ['policy_output'] = os.path.dirname(os.path.realpath(__file__))
# os.environ['policy'] = 'ebs_unattached'
# os.environ['resource_name'] = 'ocp-test'
# os.environ['user_tag_operation'] = 'read'
# os.environ['remove_tags'] = "['Manager', 'Project','Environment', 'Owner', 'Budget']"
# os.environ['username'] = 'athiruma'
# os.environ['cost_explorer_tags'] = "['User', 'Budget', 'Project', 'Manager', 'Owner', 'LaunchTime', 'Name', 'Email']"
# os.environ['file_name'] = 'tag_user.csv'
# os.environ['file_path'] = ''
# os.environ['mandatory_tags'] = "{'Budget': 'PERF-DEPT'}"
# os.environ['mandatory_tags'] = ''
# os.environ['policy'] = 'gitleaks'
# os.environ['git_access_token'] = ''
# os.environ['git_repo'] = 'https://github.com/redhat-performance'
# os.environ['several_repos'] = 'yes'
# os.environ['git_repo'] = 'https://github.com/redhat-performance/pulpperf'
# os.environ['git_repo'] = 'https://github.com/gitleakstest/gronit'
# os.environ['upload_data_elk'] = 'upload_data_elk'

log_level = os.environ.get('log_level', 'INFO').upper()
logger.setLevel(level=log_level)


def get_custodian_policies(type: str = None):
    """
    This method return a list of policies name without extension, that can filter by type
    @return: list of custodian policies name
    """
    custodian_policies = []
    # path for debug only
    # policies_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'policy')
    # policies_path working only inside the docker
    policies_path = os.path.join(os.path.dirname(__file__), 'policy', 'aws')
    for (dirpath, dirnames, filenames) in os.walk(policies_path):
        for filename in filenames:
            if not filename.startswith('__') and (filename.endswith('.yml') or filename.endswith('.py')):
                if not type:
                    custodian_policies.append(os.path.splitext(filename)[0])
                elif type and type in filename:
                    custodian_policies.append(os.path.splitext(filename)[0])
    return custodian_policies


@logger_time_stamp
@typeguard.typechecked
def run_policy(account: str, policy: str, region: str, dry_run: str):
    """
    This method run policy per region, first the custom policy and after custodian policy
    :return:
    """
    # policy Tag Cluster
    if policy == 'tag_resources':
        cluster_name = os.environ.get('resource_name', '')
        mandatory_tags = os.environ.get('mandatory_tags', {})
        tag_operation = os.environ.get('tag_operation', '')
        if mandatory_tags:
            mandatory_tags = literal_eval(mandatory_tags)  # str to dict
        if tag_operation == 'delete':
            remove_cluster_resources_tags(region=region, cluster_name=cluster_name, input_tags=mandatory_tags)
        else:
            tag_cluster_resource(cluster_name=cluster_name, mandatory_tags=mandatory_tags, region=region, tag_operation=tag_operation)
    elif policy == 'validate_iam_user_tags':
        es_host = os.environ.get('es_host', '')
        es_port = os.environ.get('es_port', '')
        es_index = os.environ.get('es_index', '')
        validate_type = os.environ.get('validate_type', '')
        user_tags = os.environ.get('user_tags', {})
        if user_tags:
            user_tags = literal_eval(user_tags)
        run_validate_iam_user_tags(es_host=es_host, es_port=es_port, es_index=es_index, validate_type=validate_type, user_tags=user_tags)
    elif policy == 'validate_cluster':
        file_path = os.environ.get('file_path', '')
        file_name = os.environ.get('file_name', '')
        file_path = file_path+file_name
        validate_zombies = ValidateZombies(file_path=file_path, region=region)
        validate_zombies.read_csv()
    elif policy == 'tag_cluster':
        cluster_name = os.environ.get('resource_name', '')
        mandatory_tags = os.environ.get('mandatory_tags', {})
        tag_operation = os.environ.get('tag_operation', '')
        if mandatory_tags:
            mandatory_tags = literal_eval(mandatory_tags)  # str to dict
        if tag_operation == 'delete':
            remove_cluster_resources_tags(region=region, cluster_name=cluster_name, input_tags=mandatory_tags, cluster_only=True)
        else:
            tag_cluster_resource(cluster_name=cluster_name, mandatory_tags=mandatory_tags, region=region, tag_operation=tag_operation, cluster_only=True )
    elif policy == 'tag_iam_user':
        user_tag_operation = os.environ.get('user_tag_operation', '')
        file_name = os.environ.get('file_name', '')
        username = os.environ.get('username', '')
        remove_keys = os.environ.get('remove_tags', '')
        if remove_keys:
            remove_keys = literal_eval(remove_keys)
        tag_iam_user(user_tag_operation=user_tag_operation, file_name=file_name, remove_keys=remove_keys, username=username)
    elif policy == 'zombie_cluster_resource':
        policy_output = os.environ.get('policy_output', '')
        resource = os.environ.get('resource', '')
        resource_name = os.environ.get('resource_name', '')
        cluster_tag = os.environ.get('cluster_tag', '')
        service_type = os.environ.get('service_type', '')
        if dry_run == 'no':  # delete
            zombie_result = zombie_cluster_resource(delete=True, region=region, resource=resource,
                                                    cluster_tag=cluster_tag, resource_name=resource_name, service_type=service_type)
        else:  # default: yes or other
            zombie_result = zombie_cluster_resource(region=region, resource=resource, cluster_tag=cluster_tag,
                                                    resource_name=resource_name, service_type=service_type)
        if policy_output:
            s3operations = S3Operations(region_name=region)
            logger.info(s3operations.save_results_to_s3(policy=policy.replace('_', '-'), policy_output=policy_output,
                                                        policy_result=zombie_result))
    elif policy == 'tag_non_cluster':
        # instance_name = os.environ['resource_name']
        mandatory_tags = os.environ.get('mandatory_tags', {})
        tag_operation = os.environ.get('tag_operation', '')
        file_name = os.environ.get('file_name', '')
        if file_name:
            file_path = os.environ.get('file_path', '')
            if file_path:
                tag_na_resources(file_name=file_name, region=region, tag_operation=tag_operation, file_path=file_path)
            else:
                tag_na_resources(file_name=file_name, region=region, tag_operation=tag_operation)
        else:
            if mandatory_tags:
                mandatory_tags = literal_eval(mandatory_tags)  # str to dict
            # mandatory_tags['Name'] = instance_name
            if tag_operation == 'delete':
                remove_tag_non_cluster_resource(mandatory_tags=mandatory_tags, region=region, dry_run='no')
            else:
                tag_non_cluster_resource(mandatory_tags=mandatory_tags, region=region, tag_operation=tag_operation)
    elif policy == 'gitleaks':
        git_access_token = os.environ.get('git_access_token')
        git_repo = os.environ.get('git_repo')
        several_repos = os.environ.get('several_repos', '')
        policy_output = os.environ.get('policy_output', '')
        try:
            if several_repos == 'yes':
                git_leaks = GitLeaks(git_access_token=git_access_token,
                                     git_repo=git_repo,
                                     several_repos=True)
            else:
                git_leaks = GitLeaks(git_access_token=git_access_token,
                                     git_repo=git_repo)
            policy_result = git_leaks.scan_repo()
            logger.info(policy_result)
            if policy_output:
                s3operations = S3Operations(region_name=region)
                logger.info(s3operations.save_results_to_s3(policy=policy, policy_output=policy_output,
                                                            policy_result=policy_result))

        except Exception as err:
            logger.exception(f'BadCredentialsException : {err}')
    # custodian policy - check if its a custodian policy
    elif any(policy == item for item in get_custodian_policies()):
        # default is dry run - change it to custodian dry run format
        if dry_run == 'yes':
            dry_run = '--dryrun'
        elif dry_run == 'no':
            dry_run = ''
        else:  # default dry run
            dry_run = '--dryrun'
        policy_output = os.environ.get('policy_output', '')
        # policies_path working only inside the docker
        # run from local - policies_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'policy', f'{policy}.yml')
        policies_path = os.path.join(os.path.dirname(__file__), 'policy', 'aws')
        if os.path.isfile(f'{policies_path}/{policy}.yml'):
            os.system(f'custodian run {dry_run} -s {policy_output} {policies_path}/{policy}.yml')
        else:
            logger.exception(f'Missing Policy name: "{policies_path}/{policy}.yml"')
            raise Exception(f'Missing Policy name: "{policies_path}/{policy}.yml"')
    else:  # local policy
        # default is dry run - change it to custodian dry run format
        if dry_run == 'yes':
            dry_run = '--dryrun'
        elif dry_run == 'no':
            dry_run = ''
        else:  # default dry run
            dry_run = '--dryrun'
        policy_output = os.environ.get('policy_output', '')
        # run from local - policies_path = os.path.join(os.path.dirname(__file__), '../' ,f'{policy}.yml')
        if os.path.isfile(policy):
            os.system(f'custodian run {dry_run} -s {policy_output} {policy}')
        else:
            logger.exception(f'Missing Policy name: {policy}')
            raise Exception(f'Missing Policy name: {policy}')


@logger_time_stamp
def main():
    """
    This main run 2 processes:
    1. ES uploader
    2. Run policy
    :return: the action output
    """
    # environment variables - get while running the docker
    region_env = os.environ.get('AWS_DEFAULT_REGION', 'us-east-2')
    dry_run = os.environ.get('dry_run', 'yes')

    account = os.environ.get('account', '')
    policy = os.environ.get('policy', '')
    upload_data_es = os.environ.get('upload_data_es', '')
    es_host = os.environ.get('es_host', '')
    es_port = os.environ.get('es_port', '')
    es_index = os.environ.get('es_index', '')
    es_doc_type = os.environ.get('es_doc_type', '')
    bucket = os.environ.get('bucket', '')

    zombie_non_cluster_polices = ['ec2_idle', 'ec2_stop', 'ebs_unattached', 'empty_buckets', 'empty_roles', 'zombie_elastic_ips', 'zombie_nat_gateways', 'zombie_snapshots']
    zombie_non_cluster_polices_runner = None
    is_zombie_non_cluster_polices_runner = policy in zombie_non_cluster_polices
    if is_zombie_non_cluster_polices_runner:
        zombie_non_cluster_polices_runner = ZombieNonClusterPolicies()

    tag_ibm_classic_infrastructure_policies = ['tag_baremetal', 'tag_vm']
    tag_ibm_classic_infrastructure_runner = None
    is_tag_ibm_classic_infrastructure_runner = policy in tag_ibm_classic_infrastructure_policies
    if is_tag_ibm_classic_infrastructure_runner:
        tag_ibm_classic_infrastructure_runner = TagResources()

    cost_explorer_policies = ['cost_explorer', 'cost_over_usage']
    cost_explorer_policies_runner = None
    is_cost_explorer_policies_runner = policy in cost_explorer_policies
    if is_cost_explorer_policies_runner:
        cost_explorer_policies_runner = CostReportPolicies()

    @logger_time_stamp
    def run_zombie_non_cluster_polices_runner():
        zombie_non_cluster_polices_runner.run()

    def run_tag_ibm_classic_infrastructure_runner():
        tag_ibm_classic_infrastructure_runner.run()

    @logger_time_stamp
    def run_cost_explorer_policies_runner():
        cost_explorer_policies_runner.run()

    # 1. ELK Uploader
    if upload_data_es:
        input_data = {'es_host': es_host,
                      'es_port': int(es_port),
                      'es_index': es_index,
                      'es_doc_type': es_doc_type,
                      'es_add_items': {'account': account},
                      'bucket': bucket,
                      'logs_bucket_key': 'logs',
                      's3_file_name': 'resources.json',
                      'region': region_env,
                      'policy': policy,
                      }
        elk_uploader = ESUploader(**input_data)
        elk_uploader.upload_to_es(account=account)
    # 2. POLICY
    elif is_zombie_non_cluster_polices_runner:
        run_zombie_non_cluster_polices_runner()
    elif is_tag_ibm_classic_infrastructure_runner:
        run_tag_ibm_classic_infrastructure_runner()
    elif is_cost_explorer_policies_runner:
        run_cost_explorer_policies_runner()
    else:
        if not policy:
            logger.exception(f'Missing Policy name: "{policy}"')
            raise Exception(f'Missing Policy name: "{policy}"')
        if region_env == 'all':
            # must be set for boto3 client default region
            # os.environ['AWS_DEFAULT_REGION'] = 'us-east-2'
            ec2 = boto3.client('ec2')
            regions_data = ec2.describe_regions()
            for region in regions_data['Regions']:
                # logger.info(f"region: {region['RegionName']}")
                os.environ['AWS_DEFAULT_REGION'] = region['RegionName']
                run_policy(account=account, policy=policy, region=region['RegionName'], dry_run=dry_run)
        else:
            run_policy(account=account, policy=policy, region=region_env, dry_run=dry_run)


main()



import json
import os
from operator import itemgetter

import pandas

from cloud_governance.common.elasticsearch.elastic_upload import ElasticUpload
from cloud_governance.common.ldap.ldap_search import LdapSearch


class CostOverUsage(ElasticUpload):

    FETCH_DAYS = 7
    COST_USAGE_DOLLAR = 1000

    def __init__(self):
        super().__init__()
        self.__ignore_mails = os.environ.get('IGNORE_MAILS', '')
        self.__ldap_host_name = os.environ.get('LDAP_HOST_NAME', '')
        self.__ldap = LdapSearch(ldap_host_name=self.__ldap_host_name)

    def get_user_used_instances(self, user_used_list: list):
        """
        This method return user used instances group by region
        @return:
        """
        region_resources = {}
        if isinstance(user_used_list, list):
            for instance in user_used_list:
                if instance:
                    if instance['Region'] in region_resources:
                        region_resources[instance['Region']].append(instance)
                    else:
                        region_resources[instance['Region']] = [instance]
        return region_resources

    def filter_sort_list(self, user_resources: dict):
        """
        This method removes duplicates from the list and return distinct instances based on user
        @param user_resources:
        @return:
        """
        for resource in user_resources:
            if 'Instances' in resource:
                if isinstance(resource['Instances'], list):
                    if resource.get('Instances')[0].get('LaunchTime'):
                        resource['Instances'] = sorted(resource['Instances'], key=itemgetter('LaunchTime'), reverse=True)
                    resource_sort_list = []
                    for item in resource['Instances']:
                        if resource_sort_list:
                            if item['InstanceId'] not in [resource_dict['InstanceId'] for resource_dict in resource_sort_list]:
                                resource_sort_list.append(item)
                        else:
                            resource_sort_list.append(item)
                    resource['Instances'] = resource_sort_list
                else:
                    resource['Instances'] = []
        return user_resources

    def aggregate_user_sum(self, data: list):
        """
        This method aggregates the es_data with User
        @param data:
        @return:
        """
        df = pandas.DataFrame.from_records(data).fillna({})
        if 'Instances' in data[0]:
            df = df.groupby('User').agg({'Cost': sum, 'Instances': sum}).reset_index()
        else:
            df = df.groupby('User').agg({'Cost': sum}).reset_index()
        return self.filter_sort_list(user_resources=df.to_dict('records'))

    def aws_user_usage(self, days: int, cost_usage: int):
        """
        This method send mail when cost_usage is greater than given cost usage in last specified days
        @param cost_usage:
        @param days:
        @return:
        """
        users = []
        cc = []
        user_data = self._elastic_search_operations.get_index_hits(days=days, index=self._es_index)
        user_data = self.aggregate_user_sum(user_data)
        for user_usage in user_data:
            user = user_usage['User']
            if user_usage['Cost'] > cost_usage:
                used_instances = self.get_user_used_instances(user_used_list=user_usage['Instances'])
                ignore_user_mails = self._literal_eval(self.__ignore_mails)
                if user not in ignore_user_mails:
                    special_user_mails = self._literal_eval(self._special_user_mails)
                    to = user if user not in special_user_mails else special_user_mails[user]
                    ldap_data = self.__ldap.get_user_details(user_name=to)
                    name = to
                    file_name = os.path.join('/tmp', f'{to}_resource.json')
                    with open(file_name, 'w') as file:
                        json.dump(used_instances, file, indent=4)
                    if ldap_data:
                        cc.append(f'{ldap_data.get("managerId")}@redhat.com')
                        name = ldap_data.get('displayName')
                    subject, body = self._mail_message.aws_user_over_usage_cost(user=to, user_usage=user_usage['Cost'], name=name, usage_cost=self.COST_USAGE_DOLLAR)
                    self._postfix_mail.send_email_postfix(subject=subject, content=body, to=to, cc=cc, filename=file_name)
                    users.append(to)
        return users

    def run(self):
        """
        This method runs the cost usage
        @return:
        """
        return self.aws_user_usage(days=self.FETCH_DAYS, cost_usage=self.COST_USAGE_DOLLAR)
